""" Tests for pyramid_amdjs.amddebug """
import mock
from pyramid_amdjs import amd, amddebug

from base import BaseTestCase


class TestAmdDirective(BaseTestCase):

    _settings = {'amd.debug': True}
    _include = False

    def test_amd_directive(self):
        self.config.include('pyramid_amdjs')

        self.assertEqual(
            'pyramid_amdjs.amddebug',
            self.config.add_amd_dir.__func__.__module__)

        self.assertEqual(
            'pyramid_amdjs.amddebug',
            self.registry[amd.ID_AMD_BUILD].__module__)

        self.assertEqual(
            'pyramid_amdjs.amddebug',
            self.registry[amd.ID_AMD_BUILD_MD5].__module__)

    def test_add_amd_dir(self):
        self.config.include('pyramid_amdjs')

        self.config.add_amd_dir('pyramid_amdjs:tests/dir/')

        path = amd.RESOLVER.resolve('pyramid_amdjs:tests/dir/').abspath()

        self.assertEqual(
            [('pyramid_amdjs:tests/dir/',path)],
            self.registry.settings['amd.debug.data']['paths'])


class TestBuildInit(BaseTestCase):

    _settings = {'amd.debug': True}

    @mock.patch('pyramid_amdjs.amddebug.build_init')
    def test_build_md5(self, m_binit):
        m_binit.return_value = '123'

        self.assertEqual(
            '202cb962ac59075b964b07152d234b70',
            amddebug.build_md5(self.request, '_'))

    @mock.patch('pyramid_amdjs.amddebug.amd')
    def test_build_init(self, m_amd):
        m_amd.build_init.return_value = '123'
        m_amd.extract_mod = amd.extract_mod

        self.config.add_static_view('_tests', 'pyramid_amdjs:tests/dir/')
        self.config.add_amd_dir('pyramid_amdjs:tests/dir/')

        res = amddebug.build_init(self.request, 'test')
        self.assertEqual('123', res)
        self.assertTrue(m_amd.build_init.called)

        mods = m_amd.build_init.call_args[0][2]
        self.assertIn(
            '"test3.css": "/_tests/test3.css?_v=6305443b362b239fad70ffc6d59c98df"',
            mods)

        self.assertIn(
            '"jca-globals": "/_tests/test.js?_v=4ce2ec81952ee8e6d0058334361babbe"',
            mods)



""" amdjs command """
from __future__ import print_function
import os
import hashlib
import argparse
import textwrap
import tempfile
from pyramid.compat import configparser, bytes_
from pyramid.path import AssetResolver
from pyramid.paster import bootstrap

from .amd import ID_AMD_MODULE, ID_AMD_SPEC, build_init
from .compat import OrderedDict, NODE_PATH, check_output
from .handlebars import ID_BUNDLE, build_hb_bundle


grpTitleWrap = textwrap.TextWrapper(
    initial_indent='* ',
    subsequent_indent='  ')

grpDescriptionWrap = textwrap.TextWrapper(
    initial_indent='    ',
    subsequent_indent='    ')


def main():
    args = AmdjsCommand.parser.parse_args()
    cmd = AmdjsCommand(args)
    cmd.run()


class AmdjsCommand(object):

    parser = argparse.ArgumentParser(description="amdjs management")
    parser.add_argument('config', metavar='config',
                        help='ini config file')

    parser.add_argument('-b', action="store_true",
                        dest='build',
                        help='Build js bundles')

    parser.add_argument('-m', action="store_true",
                        dest='amd_mods',
                        help='List amd modules')

    parser.add_argument('--deps', action="store_true",
                        dest='deps',
                        help='Print dependency tree')

    parser.add_argument('--no-min', action="store_true",
                        dest='nomin',
                        help='Do not minimize js bundles')

    def __init__(self, args):
        self.options = args
        self.env = bootstrap(args.config, options={'amd.debug': 'f'})
        self.registry = self.env['registry']
        self.request = self.env['request']
        self.resolver = AssetResolver()

    def run(self):
        if self.options.build:
            self.build_bundles()
        elif self.options.amd_mods:
            self.list_amd_mods()
        elif self.options.deps:
            self.deps_tree()
        else:
            self.parser.print_help()

    def extract_deps(self, mod):
        if mod['path']:
            path = self.resolver.resolve(mod['path']).abspath()
            if os.path.isfile(path) and path.endswith('.js'):
                text = open(path, 'rb').read().decode('utf-8')
                p1 = text.find('define(')
                if p1 >= 0:
                    p2 = text.find('function(')
                    if p2 >= 0:
                        chunk = ''.join(ch.strip()
                                        for ch in text[p1+7:p2].split())
                        if chunk.startswith("'"):
                            chunk = chunk.split(',',1)[-1]
                        deps = [d for d in
                                ''.join(ch for ch in chunk
                                        if ch not in "\"'[]").split(',')
                                if d]
                        if deps:
                            return deps

        return mod['requires']

    def deps_tree(self):
        print()
        mods = self.registry.get(ID_AMD_MODULE)
        specs = self.registry.get(ID_AMD_SPEC)
        if not specs:
            print ("No specs found")
            return

        for spec, names in specs.items():
            print(grpTitleWrap.fill('Spec: %s'%spec))

            tree = []
            seen = set()

            def process(name):
                if name in seen:
                    return

                seen.add(name)
                mod = mods.get(name)
                if mod is not None:
                    deps = self.extract_deps(mod)
                    for n in deps:
                        process(n)

                tree.append(name)

            for name in names.keys():
                if not name.endswith('.js'):
                    process(name)

            for n in tree:
                print(grpDescriptionWrap.fill(n))

            print()

    def list_amd_mods(self):
        print()
        for name, intr in sorted(self.registry.get(ID_AMD_MODULE).items()):
            print(grpTitleWrap.fill('%s: %s'%(name, intr['path'])))
            desc = grpDescriptionWrap.fill(intr['description'])
            print (desc)
            if desc:
                print()

    def build_bundles(self):
        cfg = self.registry.settings

        node_path = cfg['amd.node']
        if not node_path:
            node_path = NODE_PATH

        if not node_path:
            print ("Can't find nodejs")
            return

        if not cfg['amd.spec']:
            print ("Spec files are not specified in .ini file")
            return

        if not cfg['amd.spec-dir']:
            print ("Destination directory is not specified in .ini file")
            return

        storage = self.registry.get(ID_AMD_MODULE)
        if not storage:
            return

        resolver = self.resolver

        UGLIFY = resolver.resolve(
            'pyramid_amdjs:node_modules/uglify-js/bin/uglifyjs').abspath()

        specs = OrderedDict(cfg['amd.spec'])
        for spec, specfile in specs.items():
            print("\n\nProcessing: %s (%s)"%(spec, specfile))

            f = resolver.resolve(specfile).abspath()
            parser = configparser.SafeConfigParser()
            parser.read(f)

            # build init js
            text = build_init(self.request, spec)
            if text is not None:
                md5 = hashlib.md5()
                md5.update(text.encode('utf-8'))

                text = "var __file_hash__ = '||%s||';\n%s"%(
                    md5.hexdigest(), text)

                initpath = os.path.join(cfg['amd.spec-dir'], 'init-%s.js'%spec)
                with open(initpath, 'w') as dest:
                    dest.write(text)

            # build js bundle
            bundles = []
            processed = []

            for section in parser.sections():
                if section.endswith('.js'):
                    items = dict(parser.items(section))
                    url = items.get('url', '')
                    modules = items.get('modules', '')
                    if not modules:
                        continue

                    modules = [s for s in [s.strip() for s in modules.split()]
                               if not s.startswith('#') and s]
                    bundles.append((section, url, modules))

            tmp_storage = self.registry.get(ID_BUNDLE)

            for jsname, url, modules in bundles:
                js = []
                for module in modules:
                    if module in tmp_storage:
                        text = build_hb_bundle(
                            module, tmp_storage[module], self.registry)
                        processed.append(module)
                        js.append((module, None, text))
                        continue

                    intr = storage.get(module)
                    if not intr:
                        print ("Can't find module '%s'"%module)
                        continue

                    processed.append(module)
                    js.append((module, intr['path'],
                               resolver.resolve(intr['path']).abspath()))

                _, tpath = tempfile.mkstemp()

                print ('')
                print (grpTitleWrap.fill(jsname))

                f = open(tpath, 'ab')
                for name, path, fpath in js:
                    print(grpDescriptionWrap.fill(
                            '%s: %s'%(name, path or 'templates bundle')))

                    if path is None:
                        f.write(bytes_(fpath, 'utf8'))
                        f.write(bytes_(';\n', 'utf8'))
                    else:
                        with open(fpath, 'rb') as source:
                            f.write(source.read())
                            f.write(bytes_(';\n', 'utf8'))

                f.close()

                path = os.path.join(cfg['amd.spec-dir'], jsname)
                print ('write to:', path)
                with open(path, 'wb') as dest:
                    if self.options.nomin:
                        dest.write(open(tpath, 'rb').read())
                    else:
                        js = check_output((NODE_PATH,UGLIFY,'-nc',tpath))
                        dest.write(js)

                os.unlink(tpath)

            notprocessed = []
            for name, path in storage.items():
                if name not in processed:
                    notprocessed.append((name, path))

            if spec in ('','main') and notprocessed:
                print ("\n\nList of not processed modules:")
                for name, intr, in sorted(notprocessed):
                    print(grpDescriptionWrap.fill('%s: %s'%(name,intr['path'])))



""" assets libraries """
from pyramid_amdjs.compat import NODE_PATH


def includeme(config):
    # jquery http://jquery.org
    config.add_amd_js(
        'jquery', 'pyramid_amdjs:static/lib/jquery-1.9.0.min.js',
        'JQuery Library')
    config.add_amd_js(
        'jquery19', 'pyramid_amdjs:static/lib/jquery-1.9.0.min.js',
        'JQuery Library 1.9')

    # backbone http://backbonejs.org
    config.add_amd_js(
        'backbone', 'pyramid_amdjs:static/lib/backbone-min.js')

    # lodash https://github.com/amdjs/underscore
    config.add_amd_js(
        'underscore', 'pyramid_amdjs:static/lib/underscore-min.js')

    # json2
    config.add_amd_js(
        'json2', 'pyramid_amdjs:static/lib/json2.js')

    # moment http://momentjs.com
    config.add_amd_js(
        'moment', 'pyramid_amdjs:static/lib/moment.min.js')

    # bootstrap http://twitter.github.com/bootstrap/
    config.add_amd_js(
        'bootstrap', 'pyramid_amdjs:static/bootstrap/bootstrap.min.js',
        'Twitter bootstrap javscript library', ('jquery',))
    config.add_amd_css(
        'bootstrap-css',
        'pyramid_amdjs:static/bootstrap/bootstrap.min.css',
        'Twitter bootstrap javscript library')
    config.add_amd_css(
        'bootstrap-responsive-css',
        'pyramid_amdjs:static/bootstrap/bootstrap-responsive.min.css',
        'Twitter bootstrap javscript library (Responsive)')

    # handlebars http://handlebarsjs.com/
    node_path = config.get_settings()['amd.node']
    if not node_path:
        node_path = NODE_PATH

    if not node_path:
        config.add_amd_js(
            'handlebars', 'pyramid_amdjs:static/lib/handlebars.js',
            'Handlebars library')
    else:
        config.add_amd_js(
            'handlebars', 'pyramid_amdjs:static/lib/handlebars.runtime.js',
            'Handlebars runtime library')

    # pyramid
    config.add_amd_js(
        'pyramid', 'pyramid_amdjs:static/pyramid.js',
        'Pyramid amdjs', ('backbone'))

    # handlebars support helper
    config.add_amd_js(
        'pyramid:templates', 'pyramid_amdjs:static/templates.js',
        'Handlebars templates', ('handlebars'))

    # handlebars datetime helper
    config.add_amd_js(
        'pyramid:datetime', 'pyramid_amdjs:static/datetime.js',
        'Datetime handlebars helper', ('handlebars','moment'))



import mock
import unittest
import os, tempfile, shutil, time
from pyramid.compat import text_type, NativeIO
from pyramid.config import Configurator
from pyramid.exceptions import ConfigurationError, ConfigurationConflictError
from pyramid.httpexceptions import HTTPNotFound

from base import BaseTestCase


class TestBundleDirective(BaseTestCase):

    _include = False

    def test_amd_directive(self):
        self.assertFalse(hasattr(self.config, 'add_handlebars_bundle'))
        self.config.include('pyramid_amdjs')

        self.assertTrue(hasattr(self.config, 'add_handlebars_bundle'))

    def test_handlebars_mod(self):
        self.config.include('pyramid_amdjs')

        from pyramid_amdjs import amd

        text = amd.build_init(self.request, '_')
        self.assertIn(
            '"handlebars": "/_amdjs/static/lib/handlebars.runtime.js?_v=', text)


class TestBundleReg(BaseTestCase):

    _auto_include = False

    def test_bundle_registration(self):
        from pyramid_amdjs.handlebars import ID_BUNDLE

        self.config.add_handlebars_bundle(
            'test-bundle', 'pyramid_amdjs:tests/bundle/', i18n_domain='pyramid')
        self.config.commit()

        data = self.registry.get(ID_BUNDLE)
        self.assertIn('test-bundle', data)

        intr = data['test-bundle']
        self.assertTrue(intr['abs_path'].endswith('tests/bundle/'))
        self.assertEqual(intr['i18n_domain'], 'pyramid')

    def test_bundle_unknown(self):
        self.assertRaises(
            ConfigurationError,
            self.config.add_handlebars_bundle,
            'test-bundle', 'pyramid_amdjs:tests/unknown/')

    def test_bundle_empty_path(self):
        self.assertRaises(
            ConfigurationError,
            self.config.add_handlebars_bundle, 'test-bundle')

    def test_bundle_not_dir(self):
        self.assertRaises(
            ConfigurationError,
            self.config.add_handlebars_bundle,
            'test-bundle', 'pyramid_amdjs:tests/bundle/cat1/form.handlebars')

    def test_bundle_conflict(self):
        self.config.add_handlebars_bundle(
            'test-bundle', 'pyramid_amdjs:tests/bundle/')
        self.config.add_handlebars_bundle(
            'test-bundle', 'pyramid_amdjs:tests/bundle/')

        self.assertRaises(
            ConfigurationConflictError, self.config.commit)


class TestBundleRoute(BaseTestCase):

    def test_unknown_bundle(self):
        from pyramid_amdjs.handlebars import bundle_view

        self.request.matchdict['name'] = 'unknown'

        res = bundle_view(self.request)
        self.assertIsInstance(res, HTTPNotFound)

    def test_route(self):
        from pyramid_amdjs.handlebars import bundle_view

        self.config.add_handlebars_bundle(
            'test-bundle', 'pyramid_amdjs:tests/bundle/')
        self.request.matchdict['name'] = 'test-bundle'

        res = bundle_view(self.request)
        self.assertIn(
            '"test-bundle",["pyramid:templates","handlebars"],', res.text)
        self.assertIn(
            '"cat2":new templates.Bundle("cat2",{"form2"', res.text)

    @mock.patch('pyramid_amdjs.handlebars.log')
    def test_route_err_in_template(self, m_log):
        from pyramid_amdjs.handlebars import bundle_view

        self.config.add_handlebars_bundle(
            'test-bundle', 'pyramid_amdjs:tests/bundle3/')
        self.request.matchdict['name'] = 'test-bundle'

        bundle_view(self.request)
        arg = m_log.error.call_args[0][0]
        self.assertTrue(arg.startswith('Compilation is failed'))
        self.assertTrue(arg.endswith('tests/bundle3/form.hb'))

    def test_list_bundles(self):
        from pyramid_amdjs.handlebars import list_bundles

        self.config.add_handlebars_bundle(
            'test-bundle', 'pyramid_amdjs:tests/bundle/')

        res = list_bundles(self.request)
        self.assertEqual('test-bundle', res[0][0])
        self.assertTrue(res[0][1].startswith(
            'http://example.com/_handlebars/test-bundle.js?_v'))

    def test_list_bundles_unset(self):
        from pyramid_amdjs.handlebars import list_bundles

        config = Configurator()
        request = self.make_request()
        request.registry = config.registry

        self.assertEqual(list_bundles(request), [])

    def test_bundles_amd(self):
        from pyramid_amdjs.amd import amd_init

        self.config.add_handlebars_bundle(
            'test-bundle', 'pyramid_amdjs:tests/bundle/')

        self.request.matchdict['specname'] = '_'

        res = amd_init(self.request)
        self.assertIn('"test-bundle":"/_handlebars/test-bundle.js?_v=',res.text)

    def test_bundles_amd_spec(self):
        from pyramid_amdjs.amd import amd_init, ID_AMD_SPEC

        self.registry[ID_AMD_SPEC] = {
            'test': {'test-bundle':
                     {'name':'bundle',
                      'md5': '123',
                      'path':'pyramid_amdjs:static/example.js'}}
        }
        self.config.add_handlebars_bundle(
            'test-bundle', 'pyramid_amdjs:tests/bundle/')

        self.request.matchdict['specname'] = 'test'

        res = amd_init(self.request)
        self.assertIn(
            '"test-bundle":"/_amdjs/static/example.js?_v=123"', res.text)

    def test_build_bundle(self):
        from pyramid_amdjs.handlebars import bundle_view

        self.config.add_handlebars_bundle(
            'test-bundle', 'pyramid_amdjs:tests/bundle/')
        self.request.matchdict['name'] = 'test-bundle'

        res = bundle_view(self.request)
        self.assertIn(
            '"form-window":Handlebars.template(function', res.text)

    @mock.patch('pyramid_amdjs.handlebars.compat')
    def test_build_bundle_no_node(self, m_comp):
        from pyramid_amdjs.handlebars import bundle_view

        self.config.add_handlebars_bundle(
            'test-bundle', 'pyramid_amdjs:tests/bundle/')
        self.request.matchdict['name'] = 'test-bundle'

        m_comp.NODE_PATH = None

        res = bundle_view(self.request)
        self.assertIn(
            'form-window":Handlebars.compile("<div class=', res.text)

    @mock.patch('pyramid_amdjs.handlebars.get_localizer')
    def test_build_bundle_toplevel_i18n(self, m_loc):
        from pyramid_amdjs.handlebars import bundle_view

        cfg = self.registry.settings
        cfg['amd.tmpl-langs'] = ['en','pt_BR']

        self.config.add_handlebars_bundle(
            'test-bundle', 'pyramid_amdjs:tests/bundle2/',
            i18n_domain='pyramid')
        self.request.matchdict['name'] = 'test-bundle'

        class loc(object):
            def translate(self, t, i18n):
                if m_loc.call_args[0][0].locale_name == 'pt_BR':
                    return 'Senha'
                return t

        m_loc.return_value = loc()

        res = bundle_view(self.request)
        self.assertIn(
            "Handlebars.registerHelper('i18n-test-bundle'", res.text)
        self.assertIn(
            'var bundle=new templates.Bundle("test-bundle",{"form"', res.text)
        self.assertIn(
            'bundle.__i18n__ = {"Password": {"pt_BR": "Senha"}}', res.text)

    @mock.patch('pyramid_amdjs.handlebars.compat')
    @mock.patch('pyramid_amdjs.handlebars.get_localizer')
    def test_build_bundle_toplevel_i18n_no_nodejs(self, m_loc, m_com):
        from pyramid_amdjs.handlebars import bundle_view

        cfg = self.registry.settings
        cfg['amd.tmpl-langs'] = ['en','pt_BR']

        self.config.add_handlebars_bundle(
            'test-bundle', 'pyramid_amdjs:tests/bundle2/',
            i18n_domain='pyramid')
        self.request.matchdict['name'] = 'test-bundle'

        class loc(object):
            def translate(self, t, i18n):
                if m_loc.call_args[0][0].locale_name == 'pt_BR':
                    return 'Senha'
                return t

        m_loc.return_value = loc()
        m_com.NODE_PATH = None

        res = bundle_view(self.request)
        self.assertIn(
            "Handlebars.registerHelper('i18n-test-bundle'", res.text)
        self.assertIn(
            'var bundle=new templates.Bundle("test-bundle",{"form":Handlebars.compile(',
            res.text)
        self.assertIn(
            'bundle.__i18n__ = {"Password": {"pt_BR": "Senha"}}', res.text)


class TestBuildBundle(BaseTestCase):

    def setUp(self):
        super(TestBuildBundle, self).setUp()

        self.path = tempfile.mkdtemp()
        self.cfg = self.registry.settings

        from pyramid_amdjs import handlebars
        self.storage = self.registry.get(handlebars.ID_BUNDLE)

        self.addCleanup(shutil.rmtree, self.path)

    def test_compile_new(self):
        from pyramid_amdjs import handlebars

        self.cfg['amd.tmpl-cache'] = self.path
        prefix = os.path.split(self.path)[-1]

        f = os.path.join(self.path, 'template')
        with open(f, 'w') as fn:
            fn.write('<div>{{test}}</div>')

        tmpl = text_type(handlebars.compile_template(
            'test', f, handlebars.compat.NODE_PATH, self.path)[0])

        self.assertTrue(os.path.isfile(
                os.path.join(self.path, 'test-%s-template'%prefix)))
        self.assertTrue(os.path.isfile(
                os.path.join(self.path, 'test-%s-template.js'%prefix)))
        self.assertIn(
            'function (Handlebars,depth0,helpers,partials,data) {', tmpl)

    def test_compile_no_node(self):
        from pyramid_amdjs import handlebars

        self.cfg['amd.tmpl-cache'] = self.path
        prefix = os.path.split(self.path)[-1]

        f = os.path.join(self.path, 'template')
        with open(f, 'w') as fn:
            fn.write('<div>{{test}}</div>')

        tmpl = text_type(handlebars.compile_template(
            'test', f, None, self.path)[0])

        self.assertTrue(os.path.isfile(
                os.path.join(self.path, 'test-%s-template'%prefix)))
        self.assertTrue(os.path.isfile(
                os.path.join(self.path, 'test-%s-template.pre'%prefix)))
        self.assertIn('<div>{{test}}</div>', tmpl)

    def test_compile_new_i18n(self):
        from pyramid_amdjs import handlebars

        self.cfg['amd.tmpl-cache'] = self.path
        prefix = os.path.split(self.path)[-1]

        f = os.path.join(self.path, 'template')
        with open(f, 'w') as fn:
            fn.write('<div>{{test}}{{#i18n}}i18n text{{/i18n}}</div>')

        tmpl, i18n = handlebars.compile_template(
            'test', f, handlebars.compat.NODE_PATH, self.path)
        tmpl = text_type(tmpl)

        self.assertIn('i18n text', i18n)
        self.assertTrue(os.path.isfile(
            os.path.join(self.path, 'test-%s-template.i18n'%prefix)))
        self.assertEqual(
            open(os.path.join(self.path,'test-%s-template'%prefix),'r').read(),
            '<div>{{test}}{{#i18n-test}}i18n text{{/i18n-test}}</div>')

    def test_compile_existing(self):
        from pyramid_amdjs import handlebars

        self.cfg['amd.tmpl-cache'] = self.path
        prefix = os.path.split(self.path)[-1]

        f = os.path.join(self.path, 'template')
        with open(f, 'w') as fn:
            fn.write('<div>{{test}}</div>')

        time.sleep(0.01)

        f1 = os.path.join(self.path, 'test-%s-template'%prefix)
        with open(f1, 'w') as fn:
            fn.write('existing1')

        time.sleep(0.01)

        f2 = os.path.join(self.path, 'test-%s-template.js'%prefix)
        with open(f2, 'w') as fn:
            fn.write('existing2')

        tmpl = text_type(handlebars.compile_template(
            'test', f, handlebars.compat.NODE_PATH, self.path)[0])

        self.assertEqual('existing2', tmpl)

    def test_compile_existing_i18n(self):
        """
        Skip compilation if it is compiled already
        """
        from pyramid_amdjs import handlebars

        self.cfg['amd.tmpl-cache'] = self.path
        prefix = os.path.split(self.path)[-1]

        f = os.path.join(self.path, 'template')
        with open(f, 'w') as fn:
            fn.write('<div>{{test}}</div>')

        time.sleep(0.01)

        f1 = os.path.join(self.path, 'test-%s-template'%prefix)
        with open(f1, 'w') as fn:
            fn.write('existing1')

        time.sleep(0.01)

        f2 = os.path.join(self.path, 'test-%s-template.js'%prefix)
        with open(f2, 'w') as fn:
            fn.write('existing2')

        time.sleep(0.01)

        f1 = os.path.join(self.path, 'test-%s-template.i18n'%prefix)
        with open(f1, 'w') as fn:
            fn.write('["existing3"]')

        tmpl,i18n = handlebars.compile_template(
            'test', f, handlebars.compat.NODE_PATH, self.path)

        self.assertEqual(['existing3'], i18n)


class TestExtractI18N(unittest.TestCase):

    def test_extract(self):
        from pyramid_amdjs.handlebars import extract_i18n

        f = NativeIO('<div>{{#i18n}}Test \n message{{/i18n}}</div>')

        d = extract_i18n(f, [], [], [])
        self.assertEqual(d[0], (5, None, text_type('Test \n message'), []))


class TestNoNodeJS(BaseTestCase):

    def setUp(self):
        from pyramid_amdjs import static
        self.node_path = static.NODE_PATH

        static.NODE_PATH = None

        super(TestNoNodeJS, self).setUp()

    def tearDown(self):
        from pyramid_amdjs import static
        static.NODE_PATH = self.node_path

    def test_handlebars_mod(self):
        from pyramid_amdjs import amd

        text = amd.build_init(self.request, '_')
        self.assertIn(
            '"handlebars": "/_amdjs/static/lib/handlebars.js?_v=', text)



import mock
import os, shutil
import sys
import tempfile
from pyramid.compat import NativeIO
from pyramid_amdjs import script as amd
from pyramid_amdjs.amd import ID_AMD_MODULE, ID_AMD_SPEC

from base import BaseTestCase


class TestAmdCommand(BaseTestCase):

    @mock.patch('pyramid_amdjs.script.bootstrap')
    def test_no_params(self, m_bs):
        m_bs.return_value = {'registry': self.registry, 'request': self.request}

        sys.argv[:] = ['amdjs', 'pyramid_amdjs.ini']

        stdout = sys.stdout
        out = NativeIO()
        sys.stdout = out

        amd.main()
        sys.stdout = stdout

        val = out.getvalue()

        self.assertIn('[-h] [-b] [-m] [--deps] [--no-min] config', val)

    @mock.patch('pyramid_amdjs.script.bootstrap')
    def test_list_modules(self, m_bs):
        m_bs.return_value = {'registry': self.registry, 'request': self.request}

        self.config.add_amd_js(
            'test', 'pyramid_amdjs:tests/dir/test.js', 'Test module')

        sys.argv[1:] = ['-m', 'pyramid_amdjs.ini']

        stdout = sys.stdout
        out = NativeIO()
        sys.stdout = out

        amd.main()
        sys.stdout = stdout

        val = out.getvalue()

        self.assertIn('* test: pyramid_amdjs:tests/dir/test.js', val)
        self.assertIn('Test module', val)

    @mock.patch('pyramid_amdjs.script.build_init')
    @mock.patch('pyramid_amdjs.script.bootstrap')
    def test_build_bundle(self, m_bs, m_binit):
        m_bs.return_value = {'registry': self.registry, 'request': self.request}
        m_binit.return_value = '123'

        self.config.add_amd_js(
            'test', 'pyramid_amdjs:tests/dir/test.js', 'Test module')
        self.config.add_handlebars_bundle(
            'handlebars-test', 'pyramid_amdjs:tests/dir/', 'Handlebars bundle')

        cfg = self.registry.settings

        sys.argv[1:] = ['-b', 'pyramid_amdjs.ini']

        stdout = sys.stdout

        out = NativeIO()
        sys.stdout = out
        amd.main()
        sys.stdout = stdout

        val = out.getvalue()

        self.assertIn('Spec files are not specified in .ini file', val)

        cfg['amd.spec'] = [('main', 'pyramid_amdjs:tests/amd.spec')]

        out = NativeIO()
        sys.stdout = out
        amd.main()
        sys.stdout = stdout
        val = out.getvalue()

        self.assertIn('Destination directory is not specified in .ini file',val)

        d = tempfile.mkdtemp()
        cfg['amd.spec-dir'] = d

        out = NativeIO()
        sys.stdout = out
        amd.main()
        sys.stdout = stdout
        val = out.getvalue()

        self.assertIn('Processing: main (pyramid_amdjs:tests/amd.spec)',val)
        self.assertIn("""
* bundle.js
    test: pyramid_amdjs:tests/dir/test.js
    handlebars-test: templates bundle""", val)
        self.assertTrue(os.path.isfile(os.path.join(d, 'bundle.js')))
        self.assertFalse(os.path.isfile(os.path.join(d, 'bundle2.js')))
        self.assertTrue(os.path.isfile(os.path.join(d, 'init-main.js')))

        shutil.rmtree(d)

        d = tempfile.mkdtemp()
        cfg['amd.spec-dir'] = d

        sys.argv[1:] = ['-b', '--no-min', 'pyramid_amdjs.ini']

        out = NativeIO()
        sys.stdout = out
        amd.main()
        sys.stdout = stdout
        val = out.getvalue()

        self.assertIn('Processing: main (pyramid_amdjs:tests/amd.spec)',val)
        self.assertIn("""
* bundle.js
    test: pyramid_amdjs:tests/dir/test.js
    handlebars-test: templates bundle""", val)
        self.assertTrue(os.path.isfile(os.path.join(d, 'bundle.js')))

        shutil.rmtree(d)

    @mock.patch('pyramid_amdjs.script.bootstrap')
    def test_extract_deps(self, m_bs):
        m_bs.return_value = {'registry': self.registry, 'request': self.request}
        sys.argv[1:] = ['-b', '--no-min', 'pyramid_amdjs.ini']

        args = amd.AmdjsCommand.parser.parse_args()
        cmd = amd.AmdjsCommand(args)

        self.assertEqual(
            ['jca'],
            cmd.extract_deps({'path': 'pyramid_amdjs:tests/dir/test.js'}))

        self.assertEqual(
            ['jca'],
            cmd.extract_deps({'path': '', 'requires': ['jca']}))

    @mock.patch('pyramid_amdjs.script.bootstrap')
    def test_build_tree(self, m_bs):
        m_bs.return_value = {'registry': self.registry, 'request': self.request}
        sys.argv[1:] = ['--deps', 'pyramid_amdjs.ini']

        self.registry[ID_AMD_MODULE] = {
            'test': {'path': 'pyramid_amdjs:tests/dir/test.js'}}

        self.registry[ID_AMD_SPEC] = {'test': {'test': {}}}

        out = NativeIO()
        stdout = sys.stdout
        sys.stdout = out
        amd.main()
        sys.stdout = stdout
        val = out.getvalue()

        self.assertIn('* Spec: test', val)
        self.assertIn('jca', val)
        self.assertIn('test', val)



# pyramid_amdjs


def includeme(cfg):
    from pyramid.settings import asbool, aslist
    from pyramid.interfaces import IStaticURLInfo
    from pyramid.compat import urlparse

    from pyramid_amdjs import amd
    from pyramid_amdjs.pstatic import StaticURLInfo

    # amdjs tween
    cfg.add_tween('pyramid_amdjs.require.amdjs_tween_factory')

    def get_amdjs_data(request):
        return {'js': [], 'css': [], 'spec': '', 'fn': [], 'init': False}
    cfg.add_request_method(get_amdjs_data, 'amdjs_data', True, True)

    # static
    cfg.registry.registerUtility(StaticURLInfo(), IStaticURLInfo)

    # settings
    settings = cfg.get_settings()
    settings['amd.debug'] = asbool(settings.get('amd.debug', 't'))
    settings['amd.enabled'] = asbool(settings.get('amd.enabled', 't'))
    settings['amd.spec-dir'] = settings.get('amd.spec-dir', '').strip()
    settings['amd.tmpl-cache'] = settings.get('amd.tmpl-cache', '').strip()
    settings['amd.tmpl-langs'] = [
        s.strip() for s in aslist(settings.get('amd.tmpl-langs', ''))]
    settings['amd.node'] = settings.get('amd.node', '').strip()

    settings['static.url'] = settings.get('static.url', '').strip()
    settings['static.rewrite'] = asbool(settings.get('static.rewrite', 'f'))
    if not urlparse.urlparse(settings['static.url'])[0]:
        settings['static.rewrite'] = False
    else:
        if not settings['static.url'].endswith('/'):
            settings['static.url'] = '%s/'%settings['static.url']

    # spec settings
    specs = []
    for key, val in sorted(settings.items()):
        if key.startswith('amd.spec.'):
            specs.append((key[9:].strip(), val.strip()))

    settings['amd.spec'] = specs
    cfg.registry[amd.ID_AMD_SPEC] = {}

    # request methods
    cfg.add_request_method(amd.request_amd_init, 'init_amd')
    cfg.add_request_method(amd.request_includes, 'include_js')
    cfg.add_request_method(amd.request_css_includes, 'include_css')

    cfg.add_request_method(require.require_js, 'require_js')
    cfg.add_request_method(require.require_fn, 'require_fn')
    cfg.add_request_method(require.require_css, 'require_css')
    cfg.add_request_method(require.require_spec, 'require_spec')

    # config directives
    cfg.add_directive('add_amd_js', amd.add_js_module)
    cfg.add_directive('add_amd_css', amd.add_css_module)

    if settings['amd.debug']:
        from pyramid_amdjs import amddebug
        settings['amd.debug.data'] = {
            'paths': [], 'cache': {}, 'mods': {}}

        cfg.registry[amd.ID_AMD_BUILD] = amddebug.build_init
        cfg.registry[amd.ID_AMD_BUILD_MD5] = amddebug.build_md5
        cfg.add_directive('add_amd_dir', amddebug.add_amd_dir)
    else:
        cfg.registry[amd.ID_AMD_BUILD] = amd.build_init
        cfg.registry[amd.ID_AMD_BUILD_MD5] = amd.build_md5
        cfg.add_directive('add_amd_dir', amd.add_amd_dir)

    cfg.registry[amd.ID_AMD_MD5] = {}

    # amd init route
    cfg.add_route('pyramid-amd-init', '/_amd_{specname}.js')

    # static assets
    cfg.add_static_view('_amdjs/static', 'pyramid_amdjs:static/')

    # handlebars bundle
    from .handlebars import register_handlebars_bundle

    cfg.add_directive(
        'add_hb_bundle', register_handlebars_bundle)
    cfg.add_directive(
        'add_handlebars_bundle', register_handlebars_bundle)
    cfg.add_route(
        'pyramid-hb-bundle', '/_handlebars/{name}.js')

    # scan
    cfg.scan('pyramid_amdjs')
    cfg.include('pyramid_amdjs.static')

    # init amd specs
    amd.init_amd_spec(cfg)



""" amdjs init for request """
import logging
from collections import namedtuple
from pyramid_amdjs.amd import ID_AMD_SPEC, ID_AMD_BUILD_MD5

log = logging.getLogger('pyramid_amdjs')

amdjs_data = namedtuple('amdjs_data', 'js css spec init')


def amdjs_tween_factory(handler, registry):
    settings = registry.settings

    enabled = settings['amd.enabled']
    build_md5 = registry[ID_AMD_BUILD_MD5]
    specstorage = registry[ID_AMD_SPEC]

    def amdjs_tween(request):
        """ amdjs tween without spec support """
        if ('text/html' not in request.accept or request.is_xhr):
            return handler(request)

        response = handler(request)

        if ('amdjs_data' in request.__dict__ and
            request.amdjs_data['init'] and
            response.content_type == 'text/html' and
            response.status_code == 200 and
            response.content_length != 0 and
            response.headers.get('Transfer-Encoding', '') != 'chunked'):

            data = request.amdjs_data
            response._amdjs_inititalized = True

            spec = data['spec']
            if enabled and spec not in ('', '_'):
                specdata = specstorage.get(spec)
                if not specdata:
                    log.warn("Spec '%s' is not found.", spec)
                    spec = '_'

            if enabled and spec not in ('', '_'):
                initfile = '%s-init'%spec
                c_tmpls = [
                    '\n<script src="%s"> </script>'%(
                        request.static_url(
                            specstorage[initfile],
                            _query={'_v': build_md5(request, spec)}))]
            else:
                c_tmpls = [
                    '\n<script src="%s"> </script>'%(
                        request.route_url(
                            'pyramid-amd-init', specname='_',
                            _query={'_v': build_md5(request, '_')}))]

            mods = ["'%s'"%m for m in data['js']]
            mods.extend("'css!%s.css'"%c for c in data['css'])

            if mods:
                c_tmpls.append(
                    ('<script type="text/javascript">'
                     'curl({paths:pyramid_amd_modules},[%s])</script>' %
                     ','.join(mods)))

            if data['fn']:
                c_tmpls.append(
                    '<script type="text/javascript">\n%s</script>'%
                    ';\n'.join(data['fn']))

            text = response.text
            pos = text.find('</body>')
            if pos:
                response.text = '%s%s\n%s'%(
                    text[:pos], '\n'.join(c_tmpls), text[pos:])

        return response

    return amdjs_tween


def require_js(request, *mods):
    request.amdjs_data['init'] = True

    js = request.amdjs_data['js']
    for mod in mods:
        if mod not in js:
            js.append(mod)


def require_css(request, *mods):
    request.amdjs_data['init'] = True

    css = request.amdjs_data['css']
    for mod in mods:
        if mod not in css:
            css.append(mod)


def require_fn(request, require, fn):
    request.amdjs_data['init'] = True

    request.amdjs_data['fn'].append(
        'curl([%s], %s)'%(','.join("'%s'"%s for s in require), fn))


def require_spec(request, spec):
    request.amdjs_data['init'] = True

    if request.amdjs_data['spec']:
        log.warn("pyramid_amdjs spec is already set: %s, %s",
                 request.amdjs_data['spec'], spec)
    else:
        request.amdjs_data['spec'] = spec



import os
import logging
import hashlib
from pyramid.path import AssetResolver
from pyramid.compat import text_

from pyramid_amdjs import amd

log = logging.getLogger('pyramid_amdjs:debug')


def add_amd_dir(cfg, path):
    data = cfg.get_settings()['amd.debug.data']
    paths = data.setdefault('paths', [])

    resolver = AssetResolver()
    directory = resolver.resolve(path).abspath()

    paths.append((path, directory))
    log.info("Add resource dir: %s"%path)


def load_dir(registry, dirname, directory):
    storage = registry.settings['amd.debug.data']

    cache = storage['cache'].get(dirname)
    if cache is None:
        cache = storage['cache'][dirname] = {'mtime': 0, 'files': []}

    all_mods = storage['mods']

    # check dir mtime
    mtime = os.path.getmtime(directory)
    if os.path.getmtime(directory) > cache['mtime']:
        # unload info
        for mod, info in list(all_mods.items()):
            if 'path' in info and info['path'].startswith(directory):
                del all_mods[mod]

        cache.update(mtime=mtime, files={})

    mods = []
    for filename in os.listdir(directory):
        filepath = os.path.join(directory, filename)

        # check mtime (caching)
        try:
            mtime = os.path.getmtime(filepath)
        except OSError:
            continue

        if filename in cache['files'] and mtime <= cache['files'][filename]:
            continue

        cache['files'][filename] = mtime

        p = os.path.join(dirname, filename)

        if filename.endswith('.js'):
            with open(os.path.join(directory, filename), 'rb') as f:
                for name, deps in amd.extract_mod(
                        filename[:-3], text_(f.read()), p):
                    mods.append((name, p, amd.JS_MOD, filepath))
        if filename.endswith('.css'):
            mods.append((filename[:-4], p, amd.CSS_MOD, filepath))

    for name, fname, tp, fpath in mods:
        if name in all_mods and fpath != all_mods[name]['fpath']:
            log.error(
                "amd module '%s' already exists in '%s' file, skipping '%s'",
                name, all_mods[name]['fpath'], fpath)
        else:
            log.info("Update module information: %s path:%s", name, fname)

            md5 = hashlib.md5()
            md5.update(open(fpath, 'rb').read())
            all_mods[name] = {
                'fname': fname, 'tp': tp, 'fpath': fpath,
                'md5': md5.hexdigest()}


def build_md5(request, specname):
    initf = build_init(request, specname)

    md5 = hashlib.md5()
    md5.update(initf.encode('utf-8'))

    return md5.hexdigest()


def build_init(request, specname):
    data = request.registry.settings['amd.debug.data']

    for name, directory in data['paths']:
        load_dir(request.registry, name, directory)

    app_url = request.application_url
    app_url_len = len(app_url)

    js = []
    for name, info in data['mods'].items():
        url = '%s'%request.static_url(info['fname'], _query={'_v': info['md5']})
        if url.startswith(app_url):
            url = url[app_url_len:]

        if info['tp'] == amd.CSS_MOD:
            js.append('"%s.css": "%s"'%(name, url))
        elif info['tp'] == amd.JS_MOD:
            js.append('"%s": "%s"'%(name, url))

    return amd.build_init(request, '_', js)



""" pstatic """
from __future__ import print_function
import os
import shutil
import argparse
from pyramid.paster import bootstrap
from pyramid.path import AssetResolver
from pyramid.config.views import StaticURLInfo

ID_STATIC = 'pyramid_amdjs:static'


def main():
    args = StaticCommand.parser.parse_args()
    env = bootstrap(args.config)
    cmd = StaticCommand(args.dst, env['registry'])
    cmd.run()


class StaticCommand(object):

    parser = argparse.ArgumentParser(description="static source management")
    parser.add_argument('config', metavar='config',
                        help='ini config file')
    parser.add_argument('dst', metavar='dst',
                        help='Destination directory')

    ignores = ('~*','*~','*.py','*.pyc','*.bak','__pycache__')

    def __init__(self, dst, registry):
        self.dst = dst
        self.registry = registry

    def run(self):
        resolver = AssetResolver()
        dst = os.path.abspath(os.path.join(os.getcwd(), self.dst))

        data = self.registry.get(ID_STATIC)
        if data:
            for name, spec in data.items():
                if name.startswith('/'):
                    name = name[1:]

                dst_path = os.path.join(dst, name)
                if os.path.exists(dst_path):
                    shutil.rmtree(dst_path)

                print ('Source: %s'%spec)
                shutil.copytree(
                    resolver.resolve(spec).abspath(), dst_path,
                    ignore=shutil.ignore_patterns(*self.ignores))


class StaticURLInfo(StaticURLInfo):

    def add(self, config, name, spec, **extra):
        data = config.registry.get(ID_STATIC)
        if data is None:
            data = config.registry[ID_STATIC] = {}

        data[name] = spec

        cfg = config.get_settings()
        if cfg['static.rewrite']:
            name = cfg['static.url'] + (
                name[1:] if name.startswith('/') else name)

        super(StaticURLInfo, self).add(config, name, spec, *extra)



""" Tests for pyramid_amdjs.require module """
import mock
from pyramid.compat import text_

from base import BaseTestCase
from pyramid_amdjs.amd import RESOLVER, ID_AMD_SPEC, init_amd_spec


class TestRequire(BaseTestCase):

    def setUp(self):
        super(TestRequire, self).setUp()

        self.data = self.request.amdjs_data = {
            'js': [], 'css': [], 'spec': '', 'fn': [], 'init': False}

    def test_require_js(self):
        self.request.require_js('mod1')
        self.assertTrue(self.data['init'])
        self.assertEqual(['mod1'], self.data['js'])

    def test_require_js_unique(self):
        self.request.require_js('mod1', 'mod2', 'mod1')
        self.assertEqual(['mod1', 'mod2'], self.data['js'])

        self.request.require_js('mod1', 'mod2', 'mod1')
        self.assertEqual(['mod1', 'mod2'], self.data['js'])

    def test_require_css(self):
        self.request.require_css('mod1')
        self.assertTrue(self.data['init'])
        self.assertEqual(['mod1'], self.data['css'])

    def test_require_css_unique(self):
        self.request.require_css('mod1', 'mod2', 'mod1')
        self.assertEqual(['mod1', 'mod2'], self.data['css'])

        self.request.require_css('mod1', 'mod2', 'mod1')
        self.assertEqual(['mod1', 'mod2'], self.data['css'])

    def test_require_fn(self):
        self.request.require_fn(['jquery'], 'function($){}')
        self.assertTrue(self.data['init'])
        self.assertEqual(["curl(['jquery'], function($){})"], self.data['fn'])

    def test_require_spec(self):
        self.request.require_spec('spec')
        self.assertTrue(self.data['init'])
        self.assertEqual('spec', self.data['spec'])

    @mock.patch('pyramid_amdjs.require.log')
    def test_require_spec_mutple_times(self, m_log):
        self.request.require_spec('spec')

        self.request.require_spec('spec2')

        self.assertTrue(m_log.warn.called)
        self.assertEqual('spec', self.data['spec'])


class TestAmdjsTween(BaseTestCase):

    def setUp(self):
        super(TestAmdjsTween, self).setUp()

        self.request.accept = 'text/html'
        self.request.is_xhr = False

        self.js = []
        self.css = []
        self.fn = []
        self.spec = '_'

        self.response = self.request.response
        self.response.status_code = 200
        self.response.content_type = 'text/html; charset=utf-8'
        self.response.text = text_('<html><body><h1>Test</h1></body></html>')

    def _handler(self, request):
        if self.js:
            self.request.require_js(*self.js)
        if self.css:
            self.request.require_js(*self.js)
        for r,fn in self.fn:
            self.request.require_fn(r, fn)

        if self.spec:
            self.request.require_spec(self.spec)

        return self.response

    def _make_one(self, request):
        from pyramid_amdjs.require import amdjs_tween_factory
        return amdjs_tween_factory(self._handler, self.registry)(request)

    def test_non_200_status(self):
        self.response.status_code = 304

        res = self._make_one(self.request)
        self.assertFalse(hasattr(res, '_amdjs_inititalized'))

    def test_accept(self):
        self.request.accept = 'application/json'

        res = self._make_one(self.request)
        self.assertFalse(hasattr(res, '_amdjs_inititalized'))

    def test_xhr(self):
        self.request.is_xhr = True

        res = self._make_one(self.request)
        self.assertFalse(hasattr(res, '_amdjs_inititalized'))

    def test_response_length(self):
        self.response.content_length = 0

        res = self._make_one(self.request)
        self.assertFalse(hasattr(res, '_amdjs_inititalized'))

    def test_tween(self):
        self.js = ['mod1', 'mod2']
        self.css = ['css1']
        self.fn = [(('jquery',), 'function($){}')]
        self.spec = '_'

        res = self._make_one(self.request).text

        self.assertIn(
            '<script src="http://example.com/_amd__.js', res)
        self.assertIn(
            """<script type="text/javascript">curl({paths:pyramid_amd_modules},['mod1','mod2'])</script>""",
            res)
        self.assertIn(
            "curl(['jquery'], function($){})", res)

    @mock.patch('pyramid_amdjs.require.log')
    def test_tween_no_spec(self, m_log):
        self.registry.settings['amd.enabled'] = True
        self.spec = 'test'

        self._make_one(self.request)
        self.assertTrue(m_log.warn.called)

    def test_tween_spec(self):
        self.registry.settings['amd.enabled'] = True
        self.registry.settings['amd.spec-dir'] = RESOLVER.resolve(
                'pyramid_amdjs:tests/dir/').abspath()
        self.registry.settings['amd.spec'] = [('test', 'test.js')]
        init_amd_spec(self.config)

        self.registry[ID_AMD_SPEC] = {
            'test': {'test.js': {'path':'/test/test.js'}},
            'test-init': RESOLVER.resolve(
                'pyramid_amdjs:tests/dir/test2.js').abspath()}

        self.js = ['mod1', 'mod2']
        self.css = ['css1']
        self.fn = [(('jquery',), 'function($){}')]
        self.spec = 'test'

        res = self._make_one(self.request).text

        self.assertIn(
            '<script src="http://example.com/_amdjs/bundles/test2.js', res)
        self.assertIn(
            """<script type="text/javascript">curl({paths:pyramid_amd_modules},['mod1','mod2'])</script>""",
            res)
        self.assertIn(
            "curl(['jquery'], function($){})", res)



import os
import hashlib
import logging
from pyramid.path import AssetResolver
from pyramid.compat import configparser, text_type, string_types, text_
from pyramid.view import view_config
from pyramid.registry import Introspectable
from pyramid.response import FileResponse
from pyramid.exceptions import ConfigurationError
from pyramid.httpexceptions import HTTPNotFound

from .handlebars import list_bundles

log = logging.getLogger('pyramid_amdjs')

ID_AMD_SPEC = 'pyramid_amdjs:amd-spec'
ID_AMD_SPEC_ = 'pyramid_amdsj:amd-spec_'
ID_AMD_MODULE = 'pyramid_amdjs:amd-module'
ID_AMD_BUILD = 'pyramid_amdjs:build-init'
ID_AMD_MD5 = 'pyramid_amdjs:md5-init'
ID_AMD_BUILD_MD5 = 'pyramid_amdjs:build-md5-init'

JS_MOD = 1
CSS_MOD = 2

RESOLVER = AssetResolver()
CURL_PATH = RESOLVER.resolve('pyramid_amdjs:static/lib/curl.js').abspath()


def init_amd_spec(config, cache_max_age=None):
    cfg = config.get_settings()
    if not cfg['amd.spec']:
        return

    if not cfg['amd.spec-dir']:
        raise ConfigurationError("amd.spec-dir is required.")

    directory = RESOLVER.resolve(cfg['amd.spec-dir']).abspath()

    # register static view
    config.add_static_view('_amdjs/bundles/', directory)

    specs = {}
    for spec, specfile in cfg['amd.spec']:
        if spec in specs:
            raise ConfigurationError("Spec '%s' already defined."%spec)

        specs[spec] = specfile

    spec_mods = {}

    for spec, specfile in specs.items():
        f = RESOLVER.resolve(specfile).abspath()

        parser = configparser.SafeConfigParser()
        parser.read(f)

        mods = {}
        for section in parser.sections():
            items = dict(parser.items(section))

            if section.endswith('.js'):
                modules = [s for s in
                           [s.strip() for s in items.get('modules', '').split()]
                           if not s.startswith('#')]
                if modules:
                    md5 = hashlib.md5()
                    fpath = os.path.join(directory, section)
                    with open(fpath, 'rb') as f:
                        md5.update(f.read())

                    item = {'name': section,
                            'md5': md5.hexdigest(),
                            'path': fpath}

                mods[section] = item
                for mod in modules:
                    mods[mod] = item

        spec_mods[spec] = mods
        spec_mods['%s-init'%spec] = os.path.join(directory, 'init-%s.js'%spec)

    config.registry[ID_AMD_SPEC] = spec_mods
    config.registry[ID_AMD_SPEC_] = cache_max_age


def add_js_module(cfg, name, path, description='', requires=()):
    """ register amd js module

    :param name: name
    :param path: asset path
    :param description: module description
    :param requires: module dependencies
    """
    discr = (ID_AMD_MODULE, name)

    intr = Introspectable(ID_AMD_MODULE, discr, name, ID_AMD_MODULE)
    intr['name'] = name
    intr['path'] = path
    intr['description'] = description
    intr['tp'] = JS_MOD

    filepath = RESOLVER.resolve(path).abspath()
    md5 = hashlib.md5()
    with open(filepath, 'rb') as f:
        md5.update(f.read())

    intr['md5'] = md5.hexdigest()

    if isinstance(requires, str):
        requires = (requires,)
    intr['requires'] = requires

    storage = cfg.registry.setdefault(ID_AMD_MODULE, {})
    storage[name] = intr

    cfg.action(discr, introspectables=(intr,))
    log.info("Add js module: %s path:%s"%(name, path))


def add_css_module(cfg, name, path, description=''):
    """ register css module

    :param name: name
    :param path: asset path
    :param description: module description
    """
    discr = (ID_AMD_MODULE, name)

    intr = Introspectable(ID_AMD_MODULE, discr, name, ID_AMD_MODULE)
    intr['name'] = name
    intr['path'] = path
    intr['description'] = description
    intr['tp'] = CSS_MOD

    filepath = RESOLVER.resolve(path).abspath()
    md5 = hashlib.md5()
    with open(filepath, 'rb') as f:
        md5.update(f.read())
    intr['md5'] = md5.hexdigest()

    storage = cfg.registry.setdefault(ID_AMD_MODULE, {})
    storage[name] = intr

    cfg.action(discr, introspectables=(intr,))
    log.info("Add css module: %s path:%s"%(name, path))


def extract_mod(name, text, path):
    mods = {}

    pos = 0
    while 1:
        p1 = text.find('define(', pos)
        if p1 < 0:
            break

        p2 = text.find('function(', p1)
        if p2 < 0:
            break

        pos = p2
        chunk = ''.join(ch.strip() for ch in text[p1+7:p2].split())
        if chunk.startswith("'") or chunk.startswith('"'):
            name, chunk = chunk.split(',',1)
            name = ''.join(ch for ch in name if ch not in "\"'[]")
        else:
            log.warning("Empty name is not supported, %s.js"%name)
            continue

        deps = [d for d in
                ''.join(ch for ch in chunk
                        if ch not in "\"'[]").split(',') if d]
        mods[name] = deps

    if not mods:
        log.warning("Can't detect amdjs module name for: %s"%path)
        raise ConfigurationError("Can't detect amdjs module name for: %s"%path)

    return mods.items()


def add_amd_dir(cfg, path):
    """ read and load amd modules from directory

    :param path: asset path
    """
    directory = RESOLVER.resolve(path).abspath()

    mods = []
    for filename in os.listdir(directory):
        p = os.path.join(path, filename)

        if filename.endswith('.js'):
            with open(os.path.join(directory, filename), 'rb') as f:
                for name, deps in extract_mod(filename[:-3],text_(f.read()),p):
                    mods.append((name, p, JS_MOD))
        if filename.endswith('.css'):
            mods.append((filename[:-4], p, CSS_MOD))

    for name, p, mod in sorted(mods):
        if mod == JS_MOD:
            add_js_module(cfg, name, p)
            log.info("Add js module: %s path:%s"%(name, p))
        elif mod == CSS_MOD:
            add_css_module(cfg, name, p)
            log.info("Add css module: %s path:%s"%(name, p))

    return directory


AMD_INIT_TMPL = """
%(curl)s

var pyramid_amd_modules = {\n%(mods)s}

curl({dontAddFileExt:'.', paths: pyramid_amd_modules})
"""


def build_md5(request, specname):
    data = request.registry[ID_AMD_MD5]

    h = data.get(specname)
    if h is None:
        cfg = request.registry.settings
        specstorage = request.registry[ID_AMD_SPEC]
        specdata = specstorage.get(specname)

        initfile = '%s-init'%specname
        if specdata and cfg['amd.enabled'] and initfile in specstorage:
            with open(specstorage[initfile], 'r') as f:
                line = f.readline()
                h = line.split('||', 2)[1]
        else:
            initf = build_init(request, specname)

            md5 = hashlib.md5()
            md5.update(initf.encode('utf-8'))

            h = request.registry[ID_AMD_MD5][specname] = md5.hexdigest()

    return h


def build_init(request, specname, extra=()):
    storage = request.registry.get(ID_AMD_MODULE)

    spec_data = request.registry[ID_AMD_SPEC].get(specname)
    if spec_data is None and specname != '_':
        return None

    js = list(extra)
    if spec_data is None:
        spec_data = {}

    app_url = request.application_url
    app_url_len = len(app_url)

    spec_cache = {}

    if storage:
        for name, intr in storage.items():
            path = intr['path']
            info = spec_data.get(name)
            if info:
                url = request.static_url(
                    info['path'], _query={'_v': info['md5']})
            else:
                url = '%s'%request.static_url(path)
                if 'md5' in intr:
                    url = '%s?_v=%s'%(url, intr['md5'])

            if url.startswith(app_url):
                url = url[app_url_len:]

            if intr['tp'] == CSS_MOD:
                js.append('"%s.css": "%s"'%(name, url))
            elif intr['tp'] == JS_MOD:
                js.append('"%s": "%s"'%(name, url))

    # list handlebars bundles, in case if bundle is part of spec
    for name, url in list_bundles(request):
        info = spec_data.get(name)
        if info:
            url = request.static_url(info['path'], _query={'_v': info['md5']})

        if url.startswith(app_url):
            url = url[app_url_len:]

        js.append('"%s":"%s"'%(name, url))

    with open(CURL_PATH, 'r') as f:
        return text_type(AMD_INIT_TMPL%{
            'curl': f.read(), 'mods': ',\n'.join(sorted(js))})


@view_config(route_name='pyramid-amd-init')
def amd_init(request, **kw):
    cfg = request.registry.settings
    specstorage = request.registry[ID_AMD_SPEC]
    specname = request.matchdict['specname']
    specdata = specstorage.get(specname)
    cache_max_age = 31536000 if request.params.get('_v') else None

    initfile = '%s-init'%specname
    if specdata and cfg['amd.enabled'] and initfile in specstorage:
        return FileResponse(specstorage[initfile], request, cache_max_age)

    text = request.registry[ID_AMD_BUILD](request, specname)
    if text is None:
        return HTTPNotFound()

    response = request.response
    response.content_type = 'application/javascript'
    response.text = text

    if cache_max_age:
        response.cache_expires = cache_max_age

    return response


def request_amd_init(request, spec='', bundles=()):
    reg = request.registry
    cfg = request.registry.settings

    c_tmpls = []
    if spec and cfg['amd.enabled']:
        specstorage = reg[ID_AMD_SPEC]
        specdata = specstorage.get(spec)
        if specdata is None:
            raise RuntimeError("Spec '%s' is not found."%spec)
    else:
        spec = '_'
        specdata = ()

    if spec != '_':
        initfile = '%s-init'%spec
        c_tmpls.append(
            '<script src="%s"> </script>'%(
                request.static_url(
                    specstorage[initfile],
                    _query={'_v': reg[ID_AMD_BUILD_MD5](request, spec)})))
    else:
        c_tmpls.append(
            '<script src="%s"> </script>'%(
                request.route_url(
                    'pyramid-amd-init', specname=spec,
                    _query={'_v': reg[ID_AMD_BUILD_MD5](request, spec)})))

    for name in (bundles if not isinstance(bundles, str) else (bundles,)):
        name = '%s.js'%name
        if name in specdata:
            c_tmpls.append(
                '<script src="%s"></script>'%
                request.static_url(specdata[name]['path']))

    return '\n'.join(c_tmpls)


def request_includes(request, js=(), css=()):
    if isinstance(js, string_types):
        js = (js,)

    mods = ["'%s'"%c for c in js]

    if isinstance(css, string_types):
        css = (css,)

    mods.extend("'css!%s.css'"%c for c in css)

    return ('<script type="text/javascript">curl({paths:pyramid_amd_modules},[%s])</script>' %
            ','.join(mods))


def request_css_includes(request, css=()):
    if isinstance(css, string_types):
        css = (css,)

    return ('<script type="text/javascript">curl({paths:pyramid_amd_modules},[%s])</script>' %
            ','.join("'css!%s.css'"%c for c in css))



import sys
from pyramid import testing
from pyramid.interfaces import IRequest
from pyramid.interfaces import IRouteRequest
from pyramid.interfaces import IRequestExtensions

if sys.version_info[:2] == (2, 6):
    from unittest2 import TestCase
    TestCase
else:
    from unittest import TestCase


class BaseTestCase(TestCase):

    _include = True
    _auto_include = True
    _settings = {'amd.debug': 'f'}
    _environ = {
        'wsgi.url_scheme':'http',
        'wsgi.version':(1,0),
        'HTTP_HOST': 'example.com',
        'SCRIPT_NAME': '',
        'PATH_INFO': '/'}

    def setUp(self):
        self.init_pyramid()

    def make_request(self, environ=None, **kwargs):
        if environ is None:
            environ=self._environ
        request = testing.DummyRequest(environ=dict(environ), **kwargs)
        request.request_iface = IRequest
        request.registry = self.registry
        request._set_extensions(self.registry.getUtility(IRequestExtensions))
        return request

    def init_request_extensions(self, registry):
        from pyramid.config.factories import _RequestExtensions

        exts = registry.queryUtility(IRequestExtensions)
        if exts is None:
            exts = _RequestExtensions()
            registry.registerUtility(exts, IRequestExtensions)

    def init_pyramid(self):
        self.config = testing.setUp(
            settings=self._settings, autocommit=self._auto_include)
        self.config.get_routes_mapper()
        self.init_request_extensions(self.config.registry)
        self.registry = self.config.registry
        self.request = request = self.make_request()
        self.request.registry = self.registry

        if self._include:
            self.config.include('pyramid_amdjs')

        def set_ext():
            self.request._set_extensions(
                self.registry.getUtility(IRequestExtensions))

        self.config.action(id(self), callable=set_ext)
        self.config.begin(self.request)



""" Tests for {pyramid_amdjs.pstatic} """
import mock
from pyramid_amdjs import pstatic

from base import BaseTestCase


class TestSettingsRewrite(BaseTestCase):

    _settings = {'static.rewrite': 't',
                 'static.url': ''}

    def test_rewrite_false(self):
        """ If static.url is not set static.rewite always false """
        self.assertFalse(self.config.get_settings()['static.rewrite'])


class TestSettingsRewriteTrue(BaseTestCase):

    _settings = {'static.rewrite': 't',
                 'static.url': 'http://static.example.com'}

    def test_rewrite(self):
        """ If static.url is not set static.rewite always false """
        self.assertTrue(self.config.get_settings()['static.rewrite'])
        self.assertEqual(
            'http://static.example.com/',
            self.config.get_settings()['static.url'])


class TestStatic(BaseTestCase):

    def test_register_static_rewrite_enabled(self):
        """ Register static view, static.rewrite is enabled """
        cfg = self.config.get_settings()
        cfg['static.rewrite'] = True
        cfg['static.url'] = 'http://static.example.com/static/'

        self.config.add_static_view(
            'amdjs/static', 'pyramid_amdjs:static/')

        self.assertIn(pstatic.ID_STATIC, self.registry)

        data = self.registry[pstatic.ID_STATIC]
        self.assertIn('amdjs/static', data)

    @mock.patch('pyramid_amdjs.pstatic.os')
    @mock.patch('pyramid_amdjs.pstatic.shutil')
    def test_static_command(self, m_shutil, m_os):
        m_os.path.exists.return_value = True

        cmd = pstatic.StaticCommand('/tmp', self.registry)
        cmd.run()

        self.assertTrue(m_os.path.exists.called)
        self.assertTrue(m_shutil.rmtree.called)
        self.assertTrue(m_shutil.copytree.called)

    @mock.patch('pyramid_amdjs.pstatic.bootstrap')
    @mock.patch('pyramid_amdjs.pstatic.StaticCommand')
    def test_static_command_main(self, m_cmd, m_bootstrap):
        m_bootstrap.return_value = {'registry': self.registry}
        cmd = m_cmd.return_value = mock.Mock()
        pstatic.main()
        self.assertTrue(cmd.run.called)



import os, tempfile, mock
from pyramid.path import AssetResolver
from pyramid.compat import bytes_
from pyramid.config import Configurator
from pyramid.response import FileResponse
from pyramid.exceptions import ConfigurationError, ConfigurationConflictError
from pyramid.httpexceptions import HTTPNotFound

from base import BaseTestCase, TestCase


class TestAmdDirective(BaseTestCase):

    _include = False

    def test_amd_directive(self):
        self.assertFalse(hasattr(self.config, 'add_amd_dir'))
        self.assertFalse(hasattr(self.config, 'add_amd_js'))
        self.assertFalse(hasattr(self.config, 'add_amd_css'))
        self.config.include('pyramid_amdjs')

        self.assertTrue(hasattr(self.config, 'add_amd_dir'))
        self.assertTrue(hasattr(self.config, 'add_amd_js'))
        self.assertTrue(hasattr(self.config, 'add_amd_css'))


class TestAmd(BaseTestCase):

    _auto_include = False

    def test_js_registration(self):
        from pyramid_amdjs.amd import ID_AMD_MODULE, JS_MOD

        self.config.add_amd_js(
            'test', 'pyramid_amdjs:tests/dir/test.js')
        self.config.commit()

        data = self.registry.get(ID_AMD_MODULE)
        self.assertIn('test', data)
        self.assertEqual(data['test']['path'],
                         'pyramid_amdjs:tests/dir/test.js')
        self.assertEqual(data['test']['tp'], JS_MOD)

    def test_css_registration(self):
        from pyramid_amdjs.amd import ID_AMD_MODULE, CSS_MOD

        self.config.add_amd_css(
            'test', 'pyramid_amdjs:tests/dir/test3.css')
        self.config.commit()

        data = self.registry.get(ID_AMD_MODULE)
        self.assertIn('test', data)
        self.assertEqual(data['test']['path'],
                         'pyramid_amdjs:tests/dir/test3.css')
        self.assertEqual(data['test']['tp'], CSS_MOD)

    def test_js_registration_with_require(self):
        from pyramid_amdjs.amd import ID_AMD_MODULE

        self.config.add_amd_js(
            'test', 'pyramid_amdjs:tests/dir/test.js', requires='test2')
        self.config.commit()

        data = self.registry.get(ID_AMD_MODULE)
        self.assertEqual(data['test']['requires'], ('test2',))

    def test_reg_conflict(self):
        self.config.commit()

        self.config.add_amd_js(
            'test', 'pyramid_amdjs:tests/dir/test.js')
        self.config.add_amd_js(
            'test', 'pyramid_amdjs:tests/dir/test2.js')

        self.assertRaises(
            ConfigurationConflictError, self.config.commit)

    def test_amd_dir(self):
        from pyramid_amdjs.amd import ID_AMD_MODULE

        self.config.add_amd_dir('pyramid_amdjs:tests/dir/')
        self.config.commit()

        data = self.registry.get(ID_AMD_MODULE)
        self.assertEqual(data['jca-globals']['path'],
                         'pyramid_amdjs:tests/dir/test.js')


class TestAmdInit(BaseTestCase):

    def setUp(self):
        super(TestAmdInit, self).setUp()

        self.registry.settings['amd.enabled'] = True
        self.config.add_static_view('_tests', 'pyramid_amdjs:tests/dir/')

    @mock.patch('pyramid_amdjs.amd.build_init')
    def test_build_md5(self, m_binit):
        from pyramid_amdjs.amd import build_md5

        m_binit.return_value = '123'
        h = build_md5(self.request, '_')
        self.assertEqual('202cb962ac59075b964b07152d234b70', h)
        self.assertTrue(m_binit.called)

    @mock.patch('pyramid_amdjs.amd.build_init')
    def test_build_md5_cached(self, m_binit):
        from pyramid_amdjs.amd import build_md5, ID_AMD_MD5

        self.registry[ID_AMD_MD5]['_'] = '123'

        h = build_md5(self.request, '_')
        self.assertEqual('123', h)
        self.assertFalse(m_binit.called)

    def test_amd_init_no_spec(self):
        from pyramid_amdjs.amd import amd_init

        self.config.add_amd_js(
            'test-mod', 'pyramid_amdjs:tests/dir/test.js')

        self.request.matchdict['specname'] = 'unknown'

        resp = amd_init(self.request)
        self.assertIsInstance(resp, HTTPNotFound)

    def test_amd_init_with_v(self):
        from pyramid_amdjs.amd import amd_init, ID_AMD_SPEC

        self.request.params['_v'] = '123'
        self.request.matchdict['specname'] = 'test'

        self.registry[ID_AMD_SPEC] = \
            {'test': {'pyramid': {'name':'test',
                                  'md5': '123',
                                  'path':'pyramid_amdjs:static/example.js'}}}

        resp = amd_init(self.request)
        self.assertEqual('max-age=31536000', resp.headers['Cache-Control'])
        self.assertIn(
            '"pyramid": "/_amdjs/static/example.js?_v=123"', resp.text)

    def test_amd_init_from_file(self):
        from pyramid_amdjs.amd import amd_init, RESOLVER, ID_AMD_SPEC

        self.request.params['_v'] = '123'
        self.request.matchdict['specname'] = 'test'

        self.registry[ID_AMD_SPEC] = \
            {'test': {'pyramid': {}},
             'test-init': RESOLVER.resolve(
                'pyramid_amdjs:tests/dir/test2.js').abspath()}

        resp = amd_init(self.request)
        self.assertIsInstance(resp, FileResponse)
        self.assertEqual('max-age=31536000', resp.headers['Cache-Control'])

    def test_amd_init_with_spec_handlebars(self):
        from pyramid_amdjs.amd import amd_init, ID_AMD_SPEC

        self.request.matchdict['specname'] = 'test'
        self.request.registry = self.registry
        self.registry[ID_AMD_SPEC] = {
            'test': {'underscore':
                     {'name':'test',
                      'md5': '123',
                      'path':'pyramid_amdjs:static/example.js'}}
        }
        resp = amd_init(self.request)
        self.assertIn(
            '"underscore": "/_amdjs/static/example.js?_v=123"', resp.text)

    def test_amd_mod_paths(self):
        from pyramid_amdjs.amd import amd_init

        self.config.add_amd_js(
            'test-mod', 'pyramid_amdjs:tests/dir/test.js')
        self.config.add_amd_css(
            'test-css', 'pyramid_amdjs:tests/dir/test3.css')

        self.request.matchdict['specname'] = '_'

        resp = amd_init(self.request)
        self.assertIn('var pyramid_amd_modules = {', resp.text)
        self.assertIn(
            '"test-mod": "/_tests/test.js?_v=4ce2ec81952ee8e6d0058334361babbe"', resp.text)
        self.assertIn(
            '"test-css.css": "/_tests/test3.css?_v=6305443b362b239fad70ffc6d59c98df"', resp.text)


class TestInitAmdSpec(BaseTestCase):

    def setUp(self):
        self._files = []
        super(TestInitAmdSpec, self).setUp()

    def _create_file(self, text):
        d, fn = tempfile.mkstemp()
        self._files.append(fn)
        with open(fn, 'wb') as f:
            f.write(bytes_(text, 'utf-8'))

        return fn

    def tearDown(self):
        for f in self._files:
            os.unlink(f)

        super(TestInitAmdSpec, self).tearDown()

    def test_build_md5(self):
        from pyramid_amdjs.amd import build_md5, ID_AMD_SPEC, RESOLVER

        f = RESOLVER.resolve('pyramid_amdjs:tests/dir/test2.js').abspath()

        reg = self.registry
        reg.settings['amd.enabled'] = True
        reg[ID_AMD_SPEC]['test'] = {'test-init': f}
        reg[ID_AMD_SPEC]['test-init'] = f

        h = build_md5(self.request, 'test')
        self.assertEqual('123', h)

    def test_empty_spec(self):
        self._create_file("[test.js]\nmodules = lib1")

        cfg = self.registry.settings
        cfg['amd.spec'] = ''

        from pyramid_amdjs.amd import init_amd_spec, ID_AMD_SPEC

        # no amd-spec-dir
        init_amd_spec(self.config)

        storage = self.registry[ID_AMD_SPEC]
        self.assertEqual(storage, {})

    def test_empty_dir(self):
        fn = self._create_file("[test.js]\nmodules = lib1")

        cfg = self.registry.settings
        cfg['amd.spec'] = ['%s'%fn, 'test:%s'%fn]

        from pyramid_amdjs.amd import init_amd_spec

        # no amd-spec-dir
        self.assertRaises(ConfigurationError, init_amd_spec, self.config)

    def test_simple(self):
        fn = self._create_file("[test.js]\nmodules = lib1")

        cfg = self.registry.settings
        cfg['amd.spec'] = [('', '%s'%fn), ('test', fn)]
        cfg['amd.spec-dir'] = 'pyramid_amdjs:tests/dir/'

        from pyramid_amdjs.amd import init_amd_spec, ID_AMD_SPEC
        init_amd_spec(self.config)

        storage = self.registry[ID_AMD_SPEC]
        self.assertIn('', storage)
        self.assertTrue(storage['']['test.js']['path'].endswith('/test.js'))
        self.assertIn('test', storage)

    text1 = """
[test.js]
modules = lib1

[test.js]
modules = lib2
"""

    def test_multple_bundles(self):
        fn = self._create_file(self.text1)

        cfg = self.registry.settings
        cfg['amd.spec'] = [('test', fn), ('test', fn)]
        cfg['amd.spec-dir'] = '/unknown'

        from pyramid_amdjs.amd import init_amd_spec
        self.assertRaises(ConfigurationError, init_amd_spec, self.config)

    def test_register_static_view_for_specs(self):
        from pyramid_amdjs.amd import init_amd_spec, RESOLVER

        d = RESOLVER.resolve('pyramid_amdjs:tests/dir/').abspath()

        cfg = self.registry.settings
        cfg['amd.spec'] = [('test', 'test2.js')]
        cfg['amd.spec-dir'] = d

        m_asv = self.config.add_static_view = mock.Mock()

        init_amd_spec(self.config)
        self.assertTrue(m_asv.called)

        name, path = m_asv.call_args[0]
        self.assertEqual('_amdjs/bundles/', name)
        self.assertTrue(path.endswith('pyramid_amdjs/tests/dir'))


class TestSpecSettings(TestCase):

    def test_spec_settings(self):
        config = Configurator(settings={'amd.spec-dir': './',
                                        'amd.spec.main': '/test',
                                        'amd.spec.second': '/test1'})

        from pyramid_amdjs import includeme

        includeme(config)

        settings = config.get_settings()
        self.assertIn('amd.spec', settings)
        self.assertEqual(
            settings['amd.spec'], [('main', '/test'),('second', '/test1')])


class TestRequestRenderers(BaseTestCase):

    def setUp(self):
        super(TestRequestRenderers, self).setUp()

        from pyramid_amdjs import amd

        self.cfg = self.registry.settings
        self.registry[amd.ID_AMD_BUILD_MD5] = lambda r, s: '123'

        from pyramid.interfaces import IRequestExtensions
        extensions = self.registry.getUtility(IRequestExtensions)
        self.request._set_extensions(extensions)

    def make_request(self):
        from pyramid.request import Request
        return Request(environ=self._environ)

    def test_render_init_amd(self):
        self.cfg['amd.enabled'] = False

        text = self.request.init_amd().strip()
        self.assertEqual(
            '<script src="http://example.com/_amd__.js?_v=123"> </script>', text)

        text = self.request.init_amd('test-spec').strip()
        self.assertIn(
            '<script src="http://example.com/_amd__.js?_v=123"> </script>', text)

    def test_include_js(self):
        text = self.request.include_js('test').strip()
        self.assertIn(
            "curl({paths:pyramid_amd_modules},['test'])", text)

        text = self.request.include_js('test', 'test-css').strip()
        self.assertIn(
            "curl({paths:pyramid_amd_modules},['test','css!test-css.css'])", text)

    def test_include_css(self):
        text = self.request.include_css('test-css').strip()
        self.assertIn(
            "curl({paths:pyramid_amd_modules},['css!test-css.css'])", text)

    def test_render_js_includes_unknown_spec(self):
        self.cfg['amd.enabled'] = True

        self.assertRaises(
            RuntimeError, self.request.init_amd, 'unknown')
        self.assertRaises(
            RuntimeError, self.request.init_amd, 'spec')

    def test_render_js_includes_default(self):
        self.cfg['amd.enabled'] = True

        text = self.request.init_amd().strip()
        self.assertEqual(
            '<script src="http://example.com/_amd__.js?_v=123"> </script>', text)

    def test_render_amd_includes_spec(self):
        from pyramid_amdjs.amd import init_amd_spec, ID_AMD_SPEC, RESOLVER

        self.cfg['amd.enabled'] = True
        self.cfg['amd.spec-dir'] = RESOLVER.resolve(
                'pyramid_amdjs:tests/dir/').abspath()
        self.cfg['amd.spec'] = [('test', 'test.js')]
        init_amd_spec(self.config)

        self.registry[ID_AMD_SPEC] = {
            'test': {'test.js': {'path':'/test/test.js'}},
            'test-init': RESOLVER.resolve(
                'pyramid_amdjs:tests/dir/test2.js').abspath()}

        text = self.request.init_amd('test').strip()
        self.assertIn(
            '<script src="http://example.com/_amdjs/bundles/test2.js?_v=123"> </script>',
            text)

        m_static = self.request.static_url = mock.Mock()
        m_static.return_value = 'http://example.com/_amd_test/test.js'

        text = self.request.init_amd('test', 'test').strip()
        self.assertIn(
            '<script src="http://example.com/_amd_test/test.js"> </script>\n'
            '<script src="http://example.com/_amd_test/test.js"></script>',
            text)


class TestExtractMod(TestCase):

    def test_extract_mod(self):
        from pyramid_amdjs.amd import extract_mod

        res = extract_mod(
            'test', "define('test2', ['test3', 'test4'], function(){})", None)
        self.assertEqual(list(res), [('test2', ['test3', 'test4'])])

    @mock.patch('pyramid_amdjs.amd.log')
    def test_extract_mod_empty_name(self, log):
        from pyramid_amdjs.amd import extract_mod

        self.assertRaises(
            ConfigurationError,
            extract_mod,
            'test', "define(['test3', 'test4'], function(){})", '/test')
        self.assertEqual("Can't detect amdjs module name for: /test",
                         log.warning.call_args[0][0])

    def test_extract_mod_no_func(self):
        from pyramid_amdjs.amd import extract_mod

        self.assertRaises(
            ConfigurationError,
            extract_mod, 'test', "define('test', )", None)



import os
import sys
from setuptools import setup, find_packages

version='0.5'

install_requires = ['setuptools',
                    'pyramid >= 1.4',
]

if sys.version_info[:2] == (2, 6):
    install_requires.extend((
        'argparse',
        'ordereddict',
        'unittest2'))

if sys.version_info[:2] in ((2,6),(2,7),(3,3)):
    install_requires.extend(('simplejson',))

tests_require = install_requires + ['nose', 'mock']

def read(f):
    return open(os.path.join(os.path.dirname(__file__), f)).read().strip()


setup(name='pyramid_amdjs',
      version=version,
      description=('Pyramid JS/CSS resource management with curl.js'),
      long_description='\n\n'.join((read('README.rst'), read('CHANGES.txt'))),
      classifiers=[
          "License :: OSI Approved :: MIT License",
          "Intended Audience :: Developers",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.6",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.2",
          "Programming Language :: Python :: 3.3",
          "Programming Language :: Python :: Implementation :: CPython",
          "Framework :: Pyramid",
          "Topic :: Internet :: WWW/HTTP",
          'Topic :: Internet :: WWW/HTTP :: WSGI'],
      author='Nikolay Kim',
      author_email='fafhrd91@gmail.com',
      url='https://github.com/fafhrd91/pyramid_amdjs/',
      license='MIT',
      packages=find_packages(),
      install_requires = install_requires,
      tests_require = tests_require,
      test_suite = 'nose.collector',
      include_package_data = True,
      zip_safe = False,
      entry_points = {
          'console_scripts': [
              'amdjs = pyramid_amdjs.script:main',
              'pstatic = pyramid_amdjs.pstatic:main',
          ],
          'babel.extractors': [
              'handlebars = pyramid_amdjs.handlebars:extract_i18n',
          ],
      },
  )



import os
import shutil
import hashlib
import logging
import tempfile
from pyramid.i18n import get_localizer
from pyramid.view import view_config
from pyramid.path import AssetResolver
from pyramid.compat import text_, bytes_, text_type
from pyramid.registry import Introspectable
from pyramid.httpexceptions import HTTPNotFound
from pyramid.exceptions import ConfigurationError

from . import compat
from .compat import json, check_output

log = logging.getLogger('pyramid_amdjs')

ID_BUNDLE = 'pyramid_amdjs:handlebars'
ID_AMD_MODULE = 'pyramid_amdjs:amd-module'

HB = AssetResolver().resolve(
    'pyramid_amdjs:node_modules/handlebars/bin/handlebars').abspath()

ext_handlebars = ('.mustache', '.hb')


def register_handlebars_bundle(cfg, name,path='',description='',i18n_domain=''):
    """ Register handlebars bundle;

    :param name: module name
    :param path: asset path
    :param description:
    :param i18n_domain: i18n domain

    """
    resolver = AssetResolver()
    d = resolver.resolve(path).abspath()

    if not os.path.isdir(d):
        raise ConfigurationError("Directory is required: %s"%path)

    discr = (ID_AMD_MODULE, name)

    resolver = AssetResolver()
    abs_path = resolver.resolve(path).abspath()

    if not path or not os.path.isdir(abs_path):
        raise ConfigurationError("Directory is required: %s"%path)

    intr = Introspectable(ID_AMD_MODULE, discr, name, ID_AMD_MODULE)
    intr['name'] = name
    intr['path'] = path
    intr['abs_path'] = abs_path
    intr['description'] = description
    intr['i18n_domain'] = i18n_domain

    storage = cfg.registry.setdefault(ID_BUNDLE, {})
    storage[name] = intr

    cfg.action(discr, introspectables=(intr,))


def compile_template(name, path, node_path, cache_dir):
    tmpl = ''

    dir, tname = os.path.split(path)
    tname = os.path.join(
        cache_dir, '%s-%s-%s'%(name, os.path.split(dir)[-1], tname))

    # copy to temp dir
    if not os.path.exists(tname) or \
       (os.path.getmtime(path) > os.path.getmtime(tname)):
        shutil.copyfile(path, tname)

    i18n = []

    # check if .js file exists
    if node_path:
        cname = '%s.js'%tname
    else:
        cname = '%s.pre'%tname

    iname = '%s.i18n'%tname
    if os.path.exists(cname) and \
           (os.path.getmtime(tname) <= os.path.getmtime(cname)):
        with open(cname, 'rb') as f:
            tmpl = text_(f.read())

        if os.path.exists(iname):
            with open(iname, 'rb') as f:
                i18n.extend(json.loads(text_(f.read())))
    else:
        text = []
        with open(tname, 'rb') as f:
            data = text_(f.read())

            # i18n
            pos = 0
            for start, end, message in extract_i18n_str(data):
                text.append(data[pos:start])
                text.append('{{#i18n-%s}}'%name)
                text.append(message)
                text.append('{{/i18n-%s}}'%name)
                pos = end
                i18n.append(message)

            text.append(data[pos:])

        tmpl = bytes_(''.join(text), 'utf-8')
        with open(tname, 'wb') as f:
            f.write(tmpl)

        if i18n:
            with open(iname, 'wb') as f:
                f.write(bytes_(json.dumps(i18n), 'utf-8'))

        # compile
        if node_path:
            tmpl = check_output((node_path, HB, '-s', tname))
            if tmpl is None:
                tmpl = b''
                log.error('Compilation is failed: %s'%path)

        with open(cname, 'wb') as f:
            f.write(tmpl)

    return text_(tmpl, 'utf-8'), i18n


template = text_type("""define("%s",["pyramid:templates","handlebars"],
function(templates, Handlebars) {
var bundle=%s;templates.bundles["%s"]=bundle;%sreturn bundle})
""")

i18n_template = text_type("""\nHandlebars.registerHelper('%s',function(context, options) {return templates.i18n(bundle, this, context, options)});
bundle.__i18n__ = %s;""")

class _r(object):
    def __init__(self, registry, locale):
        self.registry = registry
        self.locale_name = locale


def build_hb_bundle(name, intr, registry):
    cfg = registry.settings

    node_path = cfg['amd.node']
    if not node_path:
        node_path = compat.NODE_PATH

    if not cfg['amd.tmpl-cache']:
        cfg['amd.tmpl-cache'] = tempfile.mkdtemp()

    cache_dir = cfg['amd.tmpl-cache']

    path = intr['abs_path']
    i18n_domain = intr['i18n_domain']

    i18n = {}
    top = []
    templates = []

    for bname in os.listdir(path):
        bdir = os.path.join(path, bname)

        if not os.path.isdir(bdir):
            if bname.endswith(ext_handlebars) and bname[0] not in ('.#~'):
                fname = os.path.join(path, bname)
                tmpl, _i18n = compile_template(name,fname,node_path,cache_dir)
                if tmpl:
                    top.append((bname.rsplit('.', 1)[0], tmpl))
                if _i18n:
                    i18n.update(dict((v, None) for v in _i18n))
        else:
            hb_tmpls = []
            for tname in os.listdir(bdir):
                if tname.endswith(ext_handlebars) and tname[0] not in ('.#~'):
                    fname = os.path.join(bdir, tname)
                    tmpl, _i18n = compile_template(
                        name, fname, node_path, cache_dir)
                    if tmpl:
                        hb_tmpls.append((tname.rsplit('.', 1)[0], tmpl))
                    if _i18n:
                        i18n.update(dict((v, None) for v in _i18n))

            if node_path:
                hb_tmpls = (
                    text_type('"%s":Handlebars.template(%s)')%(
                        name, tmpl) for name, tmpl in hb_tmpls)
            else:
                hb_tmpls = (
                    text_type('"%s":Handlebars.compile(%s)')%(
                        name, json.dumps(tmpl)) for name, tmpl in hb_tmpls)

            templates.append(
                text_type('"%s":new templates.Bundle("%s",{%s})')%(
                    bname, bname, ','.join(hb_tmpls)))

    if top:
        if node_path:
            top = (text_type('"%s":Handlebars.template(%s)')%(
                name, tmpl) for name, tmpl in top)
        else:
            top = (text_type('"%s":Handlebars.compile(%s)')%(
                name, json.dumps(tmpl)) for name, tmpl in top)

        tmpl = text_type('new templates.Bundle("%s",{%s},{%s})')%(
            name, text_type(',\n').join(top), text_type(',\n').join(templates))
    else:
        tmpl = text_type('{%s}')%(text_type(',\n').join(templates))

    name = str(name)

    if i18n:
        i18n_data = {}
        for lang in cfg['amd.tmpl-langs']:
            localizer = get_localizer(_r(registry, lang))
            for t, data in i18n.items():
                r = localizer.translate(t,i18n_domain)
                if t != r:
                    if t not in i18n_data:
                        i18n_data[t] = {}
                    i18n_data[t][lang] = r


        i18n_tmpl = i18n_template%('i18n-%s'%name, json.dumps(i18n_data))
        res = template%(name, tmpl, name, i18n_tmpl)
    else:
        res = template%(name, tmpl, name, '')

    res = text_(res, 'utf-8')

    md5 = hashlib.md5()
    md5.update(res.encode('latin-1'))

    intr['md5'] = md5.hexdigest()

    return res


@view_config(route_name='pyramid-hb-bundle')
def bundle_view(request):
    name = request.matchdict['name']
    storage = request.registry.get(ID_BUNDLE)
    if not storage or name not in storage:
        return HTTPNotFound()

    name = str(name)
    response = request.response
    response.content_type = 'application/javascript'
    response.text = build_hb_bundle(name, storage[name], request.registry)
    return response


def list_bundles(request):
    res = []

    storage = request.registry.get(ID_BUNDLE)
    if not storage:
        return res

    for name, intr in storage.items():
        build_hb_bundle(name, intr, request.registry)

        url = request.route_url(
            'pyramid-hb-bundle', name=name, _query={'_v': intr['md5']})
        res.append((name, url))

    return res


def extract_i18n_str(text):
    pos = 0
    first = last = -1

    messages = []

    while 1:
        first = text.find('{{#i18n}}', pos)
        if first >= 0:
            last = text.find('{{/i18n}}', first)
            if last >= 0:
                pos = last + 9
                messages.append((first, pos, text[first+9:last]))
                first = last = -1
                continue
        break

    return messages


def extract_i18n(fileobj, keywords, comment_tags, options):
    text = text_(fileobj.read(), 'utf-8')
    return [(first, None, message, [])
            for first, last, message in extract_i18n_str(text)]



import os
import subprocess

def check_output(*popenargs, **kwargs):
    """ python2.7 subprocess.checkoutput copy """
    process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs, **kwargs)
    output, unused_err = process.communicate()
    retcode = process.poll()
    if retcode: # pragma: nocover
        return None
    return output

try:
    import simplejson as json # faster
except ImportError: #pragma: no cover
    import json # slowest

try:
    if os.sys.platform == 'win32': # pragma: no cover
        NODE_PATH = r'C:\Program Files (x86)\nodejs\node.exe'
        os.stat(NODE_PATH)
    else:
        NODE_PATH = check_output(('which', 'node')).strip()
except: # pragma: no cover
    NODE_PATH = None

try:
    from collections import OrderedDict
except ImportError: # pragma: no cover
    from ordereddict import OrderedDict
